import logging
import typing
from datetime import datetime, timedelta

from pydantic import PrivateAttr, root_validator
try:
    from sqlalchemy import create_engine, event
    from sqlalchemy.engine.url import URL
    from sqlalchemy.orm import Session
    from sqlalchemy.pool import QueuePool, NullPool
except ImportError:
    raise ImportError("SQLAlchemyDatabase requires sqlalchemy to be installed")

from config_wrangler.config_templates.credentials import Credentials


class SQLAlchemyDatabase(Credentials):
    dialect: str
    driver: str = None
    host: str
    port: int = None
    database_name: str
    use_get_cluster_credentials: bool = False
    rs_new_credentials_seconds: int = 1800
    rs_region_name: str = None
    rs_cluster_id: str = None
    aws_access_key_id: str = None
    aws_secret_access_key: str = None
    rs_db_user_id: str = None
    rs_duration_seconds: str = 3600
    create_engine_args: typing.Dict[str, typing.Any] = {}
    arraysize: int = 5000  # Only used for Oracle
    encoding: str = None
    poolclass: str = None

    # Private attribute used to hold the redshift client
    _engine = PrivateAttr(default=None)
    _sqlite_connection = PrivateAttr(default=None)
    _rs_client = PrivateAttr(default=None)

    # Values to hide from config exports
    _private_value_atts = PrivateAttr(default={'password', 'aws_secret_access_key'})

    @root_validator(pre=True)
    def translate(cls, values):
        # Convert from old setting names to new names
        name_map = {
            'dsn': 'host',
            'dbname': 'database_name',
            'key_ring_system': 'keyring_section',
            'default_user_id': 'user_id',
        }
        for old, new in name_map.items():
            if old in values and new not in values:
                values[new] = values[old]
                del values[old]

        if values.get('dialect') in {'sqlite'}:
            values['user_id'] = 'NA'
            values['host'] = 'NA'

        if 'create_engine_args' in values:
            create_engine_args = values['create_engine_args']
            if not isinstance(create_engine_args, dict):
                create_engine_args_dict = dict()
                for arg in create_engine_args.split('\n'):
                    arg = arg.strip()
                    if arg != '':
                        arg_name, arg_value = arg.split('=')
                        arg_name = arg_name.strip()
                        arg_value = arg_value.strip()
                        try:
                            arg_value_int = int(arg_value)
                            arg_value = arg_value_int
                        except ValueError:
                            pass
                        create_engine_args_dict[arg_name] = arg_value
                values['create_engine_args'] = create_engine_args_dict

        return values

    def get_password(self) -> str:
        try:
            return super().get_password()
        except ValueError:
            if self.dialect not in {'sqlite'}:
                raise
            else:
                return None

    def get_cluster_credentials(self):
        if self._rs_client is None:
            import boto3

            self._rs_client = boto3.client(
                'redshift',
                region_name=self.rs_region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=None,
            )
        return self._rs_client.get_cluster_credentials(
                DbUser=self.rs_db_user_id,
                DbName=self.database_name,
                DurationSeconds=self.rs_duration_seconds,
                ClusterIdentifier=self.rs_cluster_id
            )

    def _get_connector(self) -> str:
        if self.driver is None:
            return self.dialect
        else:
            return f"{self.dialect}+{self.driver}"

    def get_uri(self) -> URL:
        password = self.get_password()
        if self.use_get_cluster_credentials:
            if self.aws_access_key_id is None:
                self.aws_access_key_id = self.user_id
            if self.aws_secret_access_key is None:
                self.aws_secret_access_key = password

            if self.rs_db_user_id is None:
                raise ValueError(f'rs_db_user_id required for {self.__class__.__name__} {self.host} with use_get_cluster_credentials = True.')

            credentials = self.get_cluster_credentials()
            # Overwrite the access key & secret key with the temp DB user id and password
            # Note: These go into the URL string so they need quote_plus for any special chars
            # self.user_id = quote_plus(credentials['DbUser'])
            # self.password = SecretStr(quote_plus(credentials['DbPassword']))
            self.user_id = credentials['DbUser']
            password = credentials['DbPassword']
            # TODO: save credentials[''Expiration'] for use as engine.next_get_cluster_credentials

        if self.dialect in {'sqlite'}:
            # noinspection PyTypeChecker
            self.host = None
            # noinspection PyTypeChecker
            self.user_id = None

        return URL(
            drivername=self._get_connector(),
            host=self.host,
            port=self.port,
            username=self.user_id,
            password=password,
            database=self.database_name,
        )

    def get_engine(self):
        if self._engine is None:

            kwargs = self.create_engine_args or {}
            if self.dialect == 'oracle':
                if 'arraysize' not in kwargs:
                    kwargs['arraysize'] = self.arraysize
            if self.encoding:
                kwargs['encoding'] = self.encoding
            if self.poolclass:
                if self.poolclass == 'QueuePool':
                    kwargs['poolclass'] = QueuePool
                elif self.poolclass == 'NullPool':
                    kwargs['poolclass'] = NullPool
                else:
                    raise ValueError(f'Unexpected poolclass {self.poolclass}')

            self._engine = create_engine(self.get_uri(), **kwargs)

            rs_new_credentials_timedelta = timedelta(seconds=self.rs_new_credentials_seconds)

            self._engine.next_get_cluster_credentials = datetime.now() + rs_new_credentials_timedelta

            # Make an event listener so we can get new RS credentials if needed
            @event.listens_for(self._engine, 'do_connect', named=True)
            def engine_do_connect(**kw):
                """
                listen for the 'do_connect' event
                """
                # from bi_etl.utility import dict_to_str
                # print(dict_to_str(kw))
                if self.use_get_cluster_credentials:
                    if datetime.now() > self._engine.next_get_cluster_credentials:
                        logging.info('Getting new Redshift cluster credentials')
                        new_credentials = self._rs_client.get_cluster_credentials()
                        # Note: Since we are not building a URL here we don't need/want to use quote_plus
                        kw['cparams']['user'] = new_credentials['DbUser']
                        kw['cparams']['password'] = new_credentials['DbPassword']
                        self._engine.next_get_cluster_credentials = datetime.now() + rs_new_credentials_timedelta

                # Return None to allow control to pass to the next event handler and ultimately
                # to allow the dialect to connect normally, given the updated arguments.
                return None
            # End engine_do_connect sub function

        return self._engine

    def connect(self):
        if self.dialect == 'sqlite':
            if self._sqlite_connection is None or self._sqlite_connection.closed:
                self._sqlite_connection = self.get_engine().connect()
            return self._sqlite_connection
        else:
            return self.get_engine().connect()

    def session(self, autocommit: bool = False):
        return Session(bind=self.get_engine(), autocommit=autocommit)



import inspect
import re
import typing
from configparser import RawConfigParser
from pathlib import Path

from pydantic import BaseModel
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, SHAPE_TUPLE, SHAPE_ITERABLE, SHAPE_SEQUENCE, \
    SHAPE_TUPLE_ELLIPSIS, SHAPE_SET, SHAPE_FROZENSET
from pydantic.main import ModelMetaclass
from pydantic.utils import lenient_issubclass

from config_wrangler.config_data_loaders.file_config_data_loader import FileConfigDataLoader
from config_wrangler.utils import full_name


class IniConfigDataLoader(FileConfigDataLoader):
    def _read_file(self, file_path: Path) -> typing.MutableMapping:
        self.log.debug(f"Reading {file_path}")
        self.files_read.append(file_path)
        config_data = RawConfigParser()
        config_data.read(file_path, encoding='utf8')
        config_data_dict = {section: {key: val for key, val in config_data.items(section)} for section in config_data.sections()}

        return config_data_dict

    def save_empty_config_data(self, config_model: BaseModel):
        file_path = Path(self.start_path, self.file_name)

        with file_path.open('wt') as config_file:
            logging_re = re.compile(r'typing.Dict[str, (\W+.)*LogLevel]')
            app_config_signature = inspect.signature(config_model)
            for section in app_config_signature.parameters.values():
                if section.name[0] != '_':
                    print(f"[{section.name}]", file=config_file)
                    if isinstance(section.annotation, ModelMetaclass):
                        for val in inspect.signature(section.annotation).parameters.values():
                            if val.default != val.empty:
                                print(f"; {val.name} = {val.default}", file=config_file)
                            else:
                                print(f"{val.name} = {val.annotation.__name__}_value_needed_here", file=config_file)
                    elif logging_re.match(str(section.annotation)):
                        print(f"root = INFO", file=config_file)
                        print(f"__main__=DEBUG", file=config_file)
                        print(f"requests=INFO", file=config_file)
                        print(f"; etc for each module that needs a unique log level", file=config_file)
                    elif str(section.annotation).startswith('typing.Dict[str, '):
                        print(f"; {str(section.annotation)}", file=config_file)
                        print(f"; setting1 = value", file=config_file)
                        print(f"; setting2 = value", file=config_file)
                        print(f"; etc", file=config_file)
                    else:
                        raise ValueError(f"ERROR {section} is a {section.annotation} instead of ModelMetaclass")
                print("", file=config_file)
        self.log.info(f"Created {file_path}")

    def save_config_data(self, config_data: BaseModel):
        file_path = Path(self.start_path, self.file_name)
        config_data_ini_ready = IniConfigDataLoader.prepare_config_data(config_data)

        with file_path.open('wt') as config_file:
            for section, section_data in config_data_ini_ready.items():
                for field, value in section_data:
                    print(f"{field}={value}", file=config_file)
                print("", file=config_file)
        self.log.info(f"Created {file_path}")

    @staticmethod
    def format_value_for_save(field_value):
        if isinstance(field_value, bytes):
            field_value = field_value.decode('utf8')
        else:
            field_value = str(field_value)
        return field_value

    @staticmethod
    def prepare_config_data(
            config: BaseModel,
            default_delimiter='\n',
            parents=None,
            root_config_data: typing.MutableMapping = None
    ) -> dict:
        if parents is None:
            parents = []
        config_data_dict = config.dict()
        if root_config_data is None:
            root_config_data = config_data_dict
        for field in config.__fields__.values():
            field_value = config_data_dict[field.alias]
            if field.shape == SHAPE_SINGLETON and lenient_issubclass(field.type_, BaseModel):
                section_data = IniConfigDataLoader.prepare_config_data(
                    config=getattr(config, field.alias),
                    parents=parents + [field.alias],
                    default_delimiter=default_delimiter,
                    root_config_data=root_config_data,
                )
                if len(parents) == 0:
                    root_config_data[field.alias] = section_data
                else:
                    # Flatten to 2 levels (section + field=value)
                    section_name = full_name(parents, field.alias)
                    root_config_data[section_name] = section_data

            elif (
                    field.shape in {SHAPE_LIST, SHAPE_TUPLE, SHAPE_TUPLE_ELLIPSIS, SHAPE_ITERABLE, SHAPE_SEQUENCE}
                    or field.type_ in {list, tuple}
            ):
                create_from_section_names = field.field_info.extra.get('create_from_section_names', False)
                if create_from_section_names:
                    section_name_list = []
                    for sub_section_number, sub_section_value in enumerate(getattr(config, field.alias)):
                        sub_section_id = f"{field.alias}_{sub_section_number}"
                        section_name_list.append(sub_section_id)
                        root_config_data[sub_section_id] = IniConfigDataLoader.prepare_config_data(
                            config=sub_section_value,
                            parents=None,
                            default_delimiter=default_delimiter,
                            root_config_data=root_config_data,
                        )
                    config_data_dict[field.alias] = section_name_list
                else:
                    value_list = [IniConfigDataLoader.format_value_for_save(v) for v in field_value]
                    delimiter = field.field_info.extra.get('delimiter', default_delimiter)
                    config_data_dict[field.alias] = delimiter.join(value_list)
            elif (
                    field.shape in {SHAPE_SET, SHAPE_FROZENSET}
                    or field.type_ in {set, frozenset}
            ):
                value_list = [IniConfigDataLoader.format_value_for_save(v) for v in field_value]
                delimiter = field.field_info.extra.get('delimiter', default_delimiter)
                config_data_dict[field.alias] = delimiter.join(value_list)
            else:
                # Use python format
                config_data_dict[field.alias] = IniConfigDataLoader.format_value_for_save(field_value)

        return config_data_dict



from pydantic import PrivateAttr, BaseModel


class ConfigHierarchy(BaseModel):
    _root_config = PrivateAttr()
    _parents = PrivateAttr()

    def full_item_name(self, item_name: str = None, delimiter: str = ' -> '):
        if item_name is None:
            return delimiter.join(self._parents)
        else:
            return delimiter.join(self._parents + item_name)



import typing
try:
    import boto3
except ImportError:
    raise ImportError("S3_Bucket requires boto3 to be installed")

if typing.TYPE_CHECKING:
    try:
        import botostubs
    except ImportError:
        botostubs = None

# NOTE: If you are not seeing botostubs code completion in Intellij-based IDEs,
#       please increase the intellisense filesize limit
#       e.g `idea.max.intellisense.filesize=20000` in IDE custom properties
#       (Help > Edit Custom Properties), then restart.
#       https://github.com/jeshan/botostubs#notes

from config_wrangler.config_templates.credentials import Credentials


class S3_Bucket(Credentials):
    bucket_name: str

    def get_connection(self) -> 'botostubs.S3.S3Resource':
        return boto3.resource(
            's3',
            aws_access_key_id=self.user_id,
            aws_secret_access_key=self.get_password()
        )

    def get_bucket(self, connection=None) -> 'botostubs.S3.S3Resource.Bucket':
        if connection is None:
            connection = self.get_connection()
        return connection.Bucket(self.bucket_name)




import logging
import typing
import warnings
from enum import auto

from pydantic import PrivateAttr, root_validator
from strenum import StrEnum

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy


class PasswordSource(StrEnum):
    CONFIG_FILE = auto()
    KEYRING = auto()
    KEEPASS = auto()


class Credentials(ConfigHierarchy):
    user_id: str
    password_source: PasswordSource = None
    raw_password: str = None
    keyring_section: str = 's3'
    keepass_config: str = 'keepass'
    keepass_group: str = None
    keepass_title: str = None
    validate_password_on_load: bool = True

    # Values to hide from config exports
    _private_value_atts = PrivateAttr(default={'raw_password'})

    def get_password(self) -> str:
        log = logging.getLogger(__name__)

        if self.password_source is None:
            try:
                passwords_defaults = getattr(self._root_config, 'passwords')
            except KeyError:
                raise ValueError(
                    f"{self.full_item_name()} password_source not provided and 'passwords' section not found"
                )
            try:
                self.password_source = passwords_defaults.password_source
            except AttributeError as e:
                raise ValueError(
                    f"{self.full_item_name()} password_source not provided "
                    f"and 'passwords' section does not have 'password_source' {e} "
                )

        if self.password_source == PasswordSource.KEYRING:
            if self.keyring_section is None:
                raise ValueError(f"{self.full_item_name()} keyring_section is None but password_source was keyring")
            import keyring
            password = keyring.get_password(self.keyring_section, self.user_id)
            search_info = f"{self.keyring_section} {self.user_id}"
        elif self.password_source == PasswordSource.CONFIG_FILE:
            warnings.warn(
                'Passwords stored directly in config or worse in code are not safe. Please make sure to fix this before deploying.'
            )
            password = self.raw_password
            search_info = ''
        elif self.password_source == PasswordSource.KEEPASS:
            try:
                keepass_config = getattr(self._root_config, self.keepass_config)
            except KeyError:
                raise ValueError(
                    f"{self.full_item_name()} keepass_config section {self.keepass_config} not found in config data"
                )
            try:
                get_password = keepass_config.get_password
            except AttributeError as e:
                raise ValueError(
                    f"{self.full_item_name()} keepass_config section {self.keepass_config} does not appear to be valid. "
                    f"{e}"
                )

            if self.keepass_group is None:
                if self.keyring_section is not None:
                    log.info(f"keepass_group not provided but keyring_section = {self.keyring_section}, using that for keepass group")
                    self.keepass_group = self.keyring_section
                else:
                    raise ValueError(
                        f"{self.full_item_name()} keepass_group not specified. "
                    )
            password = get_password(self.keepass_group, self.keepass_title, self.user_id)

        else:
            raise ValueError(f"{self.full_item_name()} invalid password_source {self.password_source}")

        if password is None or password == '':
            raise ValueError(f"{self.full_item_name()} password is not set. Source is {self.password_source} location = {search_info}")

        return password

    @root_validator
    def check_password(cls, values):
        if 'password_source' in values:
            if values['password_source'] is not None:
                values['password_source'] = values['password_source'].upper()

        user_id = values.get('user_id')
        if user_id == '' or user_id is None:
            if cls.__name__ == 'SQLAlchemyDatabase':
                if values.get('dialect') == 'sqlite':
                    # Bypass password checks
                    return values
            raise ValueError("user_id not provided")

        if values.get('password_source') == PasswordSource.KEYRING:
            keyring_section = values.get('keyring_section')
            if keyring_section is None:
                raise ValueError(f"{values} keyring_section is None but password_source was keyring")
        elif values.get('password_source') == PasswordSource.CONFIG_FILE:
            password = values.get('raw_password')
            if password is None or password == '':
                raise ValueError(f"{values} password_source is Config but password is not set")
        return values

    def _iter(
            self,
            to_dict: bool = False,
            by_alias: bool = False,
            include: typing.Union['AbstractSetIntStr', 'MappingIntStrAny'] = None,
            exclude: typing.Union['AbstractSetIntStr', 'MappingIntStrAny'] = None,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
    ) -> 'TupleGenerator':
        """
        Iterate through the values, but hide any password (_private_value_atts)
        values in this output.  Passwords should be directly accessed by attribute name.
        """
        for key, value in super()._iter(
            to_dict=to_dict,
            by_alias=by_alias,
            include=include,
            exclude=exclude,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        ):
            if key in self._private_value_atts and value is not None:
                value = '*' * 8
            yield key, value


class PasswordDefaults(ConfigHierarchy):
    password_source: PasswordSource = None

    @root_validator
    def check_password(cls, values):
        if 'password_source' in values:
            if values['password_source'] is not None:
                values['password_source'] = values['password_source'].upper()
        return values



import logging
import sys
import typing
from contextlib import contextmanager
from datetime import datetime
from enum import auto
from logging.handlers import RotatingFileHandler
from pathlib import Path

from CaseInsensitiveDict import CaseInsensitiveDict
from pydantic import BaseModel, ByteSize
from strenum import StrEnum

from config_wrangler.types.paths_types import AutoCreateDirectoryPath
from config_wrangler.utils import TZFormatter


class LogLevel(StrEnum):
    CRITICAL = auto()
    FATAL = auto()
    ERROR = auto()
    WARNING = auto()
    INFO = auto()
    DEBUG = auto()
    NOTSET = auto()


class LoggingConfig(BaseModel):
    console_log_level: LogLevel = LogLevel.INFO
    console_entry_format: str = '%(asctime)s - %(levelname)-8s - %(name)s: %(message)s'
    log_folder: AutoCreateDirectoryPath
    log_file_name: str = None
    log_file_name_date_time_format: str = '_%Y_%m_%d_at_%H_%M_%S'
    file_log_level: LogLevel = LogLevel.DEBUG
    log_file_entry_format: str = '%(asctime)s - %(levelname)-8s - %(name)s: %(message)s'
    log_file_max_size: ByteSize = '10 MB'
    log_files_to_keep: int = 10
    logging_date_format: str = '%Y-%m-%d %H:%M:%S%z'
    trace_logging_setup: bool = False
    log_levels: typing.Dict[str, LogLevel]

    @staticmethod
    def get_dated_log_file_name(
            log_file_prefix: str,
            date_time_format: str,
            log_file_suffix: str = '.log',
    ):
        """
        Generates a log file name with a given prefix, suffix and date time format.

        Parameters
        ----------
        log_file_prefix: str
            The part of the log file name before the date
        log_file_suffix: str
            The part of the log file name after the date
        date_time_format: str
            Optional. The date time format to use. Defaults to '_%Y_%m_%d_at_%H_%M_%S'
        """
        return f"{log_file_prefix}{datetime.now().strftime(date_time_format)}{log_file_suffix}"

    def add_log_file_handler(
            self,
            log_file_prefix: str = None,
            add_date_to_log_file_name: bool = False,
            log_file_suffix: str = '.log',
    ) -> logging.Handler:
        log = logging.getLogger(__name__)
        root_logger = logging.getLogger()
        if log_file_prefix is None:
            if self.log_file_name is not None:
                log_file_name = self.log_file_name
            else:
                log_file_name = None
        else:
            if add_date_to_log_file_name:
                log_file_name = LoggingConfig.get_dated_log_file_name(
                    log_file_prefix=log_file_prefix,
                    date_time_format=self.log_file_name_date_time_format,
                    log_file_suffix=log_file_suffix,
                )
            else:
                log_file_name = f"{log_file_prefix}.{log_file_suffix}"

        file_handler = None
        if log_file_name is not None:
            if self.log_folder is not None and not Path(log_file_name).is_absolute():
                log_file_path = Path(self.log_folder, log_file_name)
            else:
                log_file_path = Path(log_file_name)

            if self.trace_logging_setup:
                log.info('Logging path = {}'.format(log_file_path))

            # Setup file logging

            # Make sure the directory exists
            dir_name = log_file_path.parent
            if dir_name.is_absolute():
                dir_name.mkdir(parents=True, exist_ok=True)

            file_handler = RotatingFileHandler(filename=log_file_path,
                                               maxBytes=self.log_file_max_size,
                                               backupCount=self.log_files_to_keep,
                                               encoding='utf8',
                                               )
            log_file_entry_formatter = TZFormatter(self.log_file_entry_format, self.logging_date_format)
            file_handler.setFormatter(log_file_entry_formatter)
            file_handler.setLevel(self.file_log_level)
            root_logger.addHandler(file_handler)
            log.info('File log level = {}'.format(self.file_log_level))
        else:
            if self.trace_logging_setup:
                log.info('No log filename defined. File logging skipped.')
        return file_handler

    def remove_log_handler(self, handler: logging.Handler):
        log = logging.getLogger(__name__)
        if handler is not None:
            if self.trace_logging_setup:
                log.info(f"Closing log handler {handler}")
            root_logger = logging.getLogger()
            root_logger.removeHandler(handler)
            handler.close()
        else:
            if self.trace_logging_setup:
                log.info('No handler provided to remove_log_handler')

    @contextmanager
    def log_file_manager(self, file_name: str):
        log_file_handler = self.add_log_file_handler(log_file_name=file_name)
        try:
            yield log_file_handler
        finally:
            # Code to release resource, e.g.:
            self.remove_log_handler(log_file_handler)

    def setup_log_levels(self):
        configured_loggers = dict()
        root_logger = logging.getLogger()

        # Monkey-patch getLogger's dict to be case-insensitive        
        logger_dict = CaseInsensitiveDict(logging.Logger.manager.loggerDict)  # @UndefinedVariable
        logging.Logger.manager.loggerDict = logger_dict

        for logger_class, desired_level_name in self.log_levels.items():
            if logger_class.lower() == 'root':
                logger = root_logger
            else:
                logger = logging.getLogger(logger_class)
            configured_loggers[logger_class] = logger
            logger.propagate = True
            if desired_level_name is None:
                desired_level_name = 'INFO'
            else:
                desired_level_name = desired_level_name.upper()
            if self.trace_logging_setup:
                # Note: We can't use logging here yet
                print(f'Setting logger {logger.name} to {desired_level_name}')
            logger.setLevel(desired_level_name)

        # Will not include root logger
        for logger_class in sorted(logging.Logger.manager.loggerDict):
            logger = logging.getLogger(logger_class)
            if self.trace_logging_setup:
                print(f'Logger {logger_class} handlers {logger.handlers}')
            if logger_class not in configured_loggers:
                if self.trace_logging_setup:
                    # Note: We can't use logging here yet
                    print(
                        f"Checking existing logger {logger_class} "
                        f"level {logging.getLevelName(logger.getEffectiveLevel())}"
                    )
                for compare_logger in sorted(configured_loggers):
                    if logger_class.startswith(compare_logger):
                        parent_logger = configured_loggers[compare_logger]
                        level = parent_logger.getEffectiveLevel()
                        logger.setLevel(level)
                        if self.trace_logging_setup:
                            # Note: We can't use logging here yet
                            print(
                                f"Existing logger {logger_class} re-setup with {parent_logger.name} "
                                f"settings {logging.getLevelName(level)}"
                            )
                        logger.propagate = True

        if self.trace_logging_setup:
            # Note: We can't use logging here yet
            print(f"Root logging level is {logging.getLevelName(root_logger.getEffectiveLevel())}")
            print(f"Root logging handlers are {root_logger.handlers}")

    def setup_logging(
            self,
            log_file_prefix: str = None,
            add_date_to_log_file_name: bool = False,
            log_file_suffix: str = '.log',
            console_output=None,
            use_log_file_setting=True,
    ):
        """
        Setup logging based on configuration.
        """

        root_logger = logging.getLogger()

        if not hasattr(root_logger, 'config_wrangler_setup_done'):
            # Store info that we have already setup logging in the root logger.
            # This could be stored in the config, it would be too easy for methods to open
            # their own configs and thus will not know the logging had already been setup.
            root_logger.config_wrangler_setup_done = True

            # Close out any existing handlers
            for handler in root_logger.handlers:
                handler.flush()
                handler.close()

            # Reset the handlers
            root_logger.handlers.clear()

            if self.trace_logging_setup:
                # Note: We can't use logging here yet
                print('logging.trace_setup is True}')
                print(f'Starting root logger handlers {root_logger.handlers}')

            # Modify the root logger level to at least INFO for now. setup_log_levels could change it
            root_logger.setLevel(logging.INFO)

            if console_output is None:
                error_output = sys.stderr
                regular_output = sys.stdout
            else:
                error_output = console_output
                regular_output = console_output

            console_error_log = logging.StreamHandler(error_output)
            console_error_log.setLevel(logging.ERROR)
            root_logger.addHandler(console_error_log)

            console_log = logging.StreamHandler(regular_output)
            console_log.setLevel(self.console_log_level)

            # define a filter for non-error messages
            def non_error(record):
                return record.levelno != logging.ERROR

            # Errors go to console_error_log above so we don't want them here as well
            console_log.addFilter(non_error)

            root_logger.addHandler(console_log)

            console_entry_format = self.console_entry_format
            if console_entry_format:
                console_entry_formater = TZFormatter(console_entry_format, self.logging_date_format)
                console_log.setFormatter(console_entry_formater)
                console_error_log.setFormatter(console_entry_formater)

            self.setup_log_levels()

            # Switch to this modules logger
            log = logging.getLogger(__name__)

            logging.captureWarnings(True)

            log_level_name = logging.getLevelName(log.getEffectiveLevel())
            if self.trace_logging_setup:
                log.info('This modules logging level is {}'.format(log_level_name))

            if use_log_file_setting or log_file_prefix is not None:
                return self.add_log_file_handler(
                    log_file_prefix=log_file_prefix,
                    add_date_to_log_file_name=add_date_to_log_file_name,
                    log_file_suffix=log_file_suffix,
                )
            else:
                if self.trace_logging_setup:
                    log.info('use_log_file_setting = False. setup_log_file not called.')
                return None



import typing
from pathlib import Path

from pydantic import PrivateAttr, root_validator

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_templates.credentials import Credentials, PasswordSource


class Keepass(ConfigHierarchy):
    database_path: Path
    password_source: PasswordSource = PasswordSource.KEYRING
    raw_password: str = None
    keyring_section: str = 's3'
    keyring_user_id: str = None
    alternate_group_names: typing.Dict[str, str] = {}

    _db = PrivateAttr(default=None)
    _alternate_group_names_lower = PrivateAttr(default=None)

    @root_validator
    def check_password(cls, values):
        if 'password_source' in values:
            if values['password_source'] is not None:
                values['password_source'] = values['password_source'].upper()
        return values

    def open_database(self) -> 'pykeepass.PyKeePass':
        if self._db is None:
            from pykeepass import PyKeePass

            credentials_args = dict(**self.__dict__)
            credentials_args['user_id'] = credentials_args.get('keyring_user_id')
            credentials = Credentials(**credentials_args)
            password = credentials.get_password()
            try:
                self._db = PyKeePass(self.database_path, password=password)
            except Exception as e:
                raise ValueError(f"PyKeePass error from {self.database_path}: {repr(e)}")

            self._alternate_group_names_lower = {k.lower(): v.lower() for k, v in self.alternate_group_names.items()}
        return self._db

    def get_password(self, group: str, title: str, user_id: str):
        kp = self.open_database()

        group_matches = list()
        group_lower = group.lower()
        group_search_set = {group_lower}
        alt_group_lower = self._alternate_group_names_lower.get(group_lower)
        if alt_group_lower is not None:
            group_search_set.add(alt_group_lower)

        for group_object in kp.groups:
            if group_object.name is not None and group_object.name.lower() in group_search_set:
                group_matches.append(group_object)
        if len(group_matches) == 0:
            raise ValueError(f"Keepass group '{group}' not found in {self.database_path}")
        elif len(group_matches) > 1:
            raise ValueError(
                f"Keepass group '{group}' multiple matches in {self.database_path}.  group_matches: {group_matches}"
            )
        else:
            group_object = group_matches[0]
            selectors = []
            if title is not None:
                selectors.append(f"Title = '{title}'")
                title_lower = title.lower()
            else:
                title_lower = None

            if user_id is not None:
                selectors.append(f"User_name = '{user_id}'")
                user_id_lower = user_id.lower()
            else:
                user_id_lower = None

            selectors_str = ' and '.join(selectors)
            entry_matches = list()
            for entry in group_object.entries:
                match = True
                if title is not None:
                    if entry.title is not None and entry.title.lower() == title_lower:
                        match = True
                    else:
                        match = False
                if user_id is not None:
                    if entry.username is not None and entry.username.lower() == user_id_lower:
                        match = True
                    else:
                        match = False
                if match:
                    entry_matches.append(entry.password)
            if len(entry_matches) == 0:
                raise ValueError(f"Keepass group '{group_object.name}' does not have entry for {selectors_str} in {self.database_path}")
            elif len(entry_matches) > 1:
                selectors_str = ' and '.join(selectors)
                raise ValueError(
                    f"Keepass group '{group_object.name}' multiple matches for entry for {selectors_str} in {self.database_path}"
                )
            else:
                entry_password = entry_matches[0]
                return entry_password



import ast
import json
import logging
import re
import typing
from datetime import timezone, datetime

from pydantic import BaseModel
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, SHAPE_TUPLE, SHAPE_ITERABLE, SHAPE_SEQUENCE, \
    SHAPE_TUPLE_ELLIPSIS, SHAPE_SET, SHAPE_FROZENSET, SHAPE_DICT, SHAPE_DEFAULTDICT, Field
from pydantic.utils import lenient_issubclass

from config_wrangler.config_exception import ConfigError


class TZFormatter(logging.Formatter):
    local_timezone = datetime.now(timezone.utc).astimezone().tzinfo

    @staticmethod
    def tz_aware_converter(timestamp) -> datetime:
        return datetime.fromtimestamp(timestamp, TZFormatter.local_timezone)

    def formatTime(self, record, datefmt=None):
        dt = TZFormatter.tz_aware_converter(record.created)
        if datefmt is not None:
            result = dt.strftime(datefmt)
        else:
            time_part = dt.strftime(self.default_time_format)
            result = self.default_msec_format % (time_part, record.msecs)
        return result


def merge_configs(child: typing.MutableMapping, parent: typing.MutableMapping) -> None:
    for section in parent:
        if section not in child:
            child[section] = parent[section]
        else:
            if isinstance(child[section], typing.MutableMapping):
                merge_configs(child[section], parent[section])
            else:
                pass


def resolve_variable(root_config_data: typing.MutableMapping, variable_name: str, default=None, part_delimiter=':') -> typing.Any:
    variable_name_parts = variable_name.split(part_delimiter)
    result = root_config_data
    for part in variable_name_parts:
        if part in result:
            result = result[part]
        else:
            return default
    return result


_interpolation_re = re.compile(r"\${([^}]+)}")


def interpolate_values(container: typing.MutableMapping, root_config_data: typing.MutableMapping):
    for section in container:
        value = container[section]
        if isinstance(value, typing.MutableMapping):
            interpolate_values(value, root_config_data=root_config_data)
        elif isinstance(value, str):
            if '$' in value:
                variables_cnt = 0
                result_values = []
                next_start = 0
                for variable_found in _interpolation_re.finditer(value):
                    variables_cnt += 1
                    variable_name = variable_found.groups()[0]
                    variable_text = f"${{{variable_name}}}"
                    var_start, var_end = variable_found.span()

                    if ':' in variable_name:
                        variable_replacement = resolve_variable(
                            root_config_data, variable_name, default=variable_text, part_delimiter=':'
                        )
                    elif '.' in variable_name:
                        variable_replacement = resolve_variable(
                            root_config_data, variable_name, default=variable_text, part_delimiter='.'
                        )
                    else:
                        variable_replacement = container.get(variable_name, default=variable_text)

                    result_values.append(value[next_start:var_start])
                    result_values.append(variable_replacement)
                    next_start = var_end + 1
                if variables_cnt > 0:
                    if next_start < len(value):
                        result_values.append(value[next_start:])
                    container[section] = ''.join(result_values)


def parse_as_literal_or_json(value: str) -> typing.Any:
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError) as el:
        try:
            return json.loads(value)
        except ValueError as ej:
            raise ConfigError(
                f"Value {value} could not be parsed as python literal '{el}' or json '{ej}'"
            )


def parse_delimited_list(field: Field, value: str) -> typing.Sequence:
    value = value.strip()
    if len(value) == 0:
        return []
    delimiter = field.field_info.extra.get('delimiter', None)
    if delimiter is None and value[0] not in {'[', '{'}:
        # Try to automatically recognize the delimiter
        for delimiter in [',', '\n', '|']:
            if delimiter in value:
                break
    if delimiter is not None:
        if value[0] == delimiter:
            value = value[1:]
        result = [v.strip() for v in value.split(delimiter)]
        if field.type_ == int:
            result = [int(v) for v in result]
        elif field.type_ == float:
            result = [float(v) for v in result]
    else:
        try:
            result = parse_as_literal_or_json(value)
        except ValueError as e:
            raise ConfigError(f"Field {field.alias} {e}")
    return result


def full_name(parents: typing.List[str], field_name: str):
    return '.'.join(parents + [field_name])


def match_config_data_to_model(model: BaseModel, config_data: typing.MutableMapping, root_config_data: typing.MutableMapping = None, parents=None):
    if parents is None:
        parents = []
    if root_config_data is None:
        root_config_data = config_data

    # Make mappings from lower case names to actual config field names
    config_name_map = {key.lower(): key for key in config_data}
    root_config_name_map = {key.lower(): key for key in root_config_data}

    # Scan all model fields to look for values in the MutableMapping
    for field in model.__fields__.values():
        # Check for either a direct name match or a case in-sensitive match
        field_lower = field.alias.lower()
        if field.alias in config_data:
            found = True
        elif field_lower in config_name_map:
            found = True
            field_name = config_name_map[field_lower]
            config_data[field.alias] = config_data[field_name]
            del config_data[field_name]
        else:
            found = False

        # Check for nested objects set using top level dotted names (e.g. [parent.child])
        # (either a direct name match or a case in-sensitive match)
        if not found and len(parents) > 0:
            section_name = '.'.join(parents + [field.alias])
            section_name_lower = section_name.lower()
            if section_name in root_config_data:
                found = True
                config_data[field.alias] = root_config_data[section_name]
            elif section_name_lower in root_config_name_map:
                found = True
                section_name = root_config_name_map[section_name_lower]
                config_data[field.alias] = root_config_data[section_name]

        if found:
            # Check if we should recurse into deeper structures or parse strings into structures
            value = config_data[field.alias]
            if lenient_issubclass(field.type_, BaseModel) or hasattr(field.type_, '__fields__'):
                create_from_section_names = field.field_info.extra.get('create_from_section_names', False)
                if create_from_section_names:
                    inherit = field.field_info.extra.get('inherit', False)
                    if isinstance(value, str):
                        section_names = parse_delimited_list(field, value)
                    else:
                        section_names = value
                    section_contents = []
                    for section in section_names:
                        if isinstance(section, str):
                            if section in root_config_data:
                                if inherit:
                                    for inherit_key, inherit_value in config_data.items():
                                        if inherit_key not in root_config_data[section]:
                                            root_config_data[section][inherit_key] = inherit_value
                                match_config_data_to_model(field.type_, root_config_data[section], root_config_data=root_config_data, parents=parents + [field.alias])
                                section_contents.append(root_config_data[section])
                            elif section in config_data:
                                if inherit:
                                    for inherit_key, inherit_value in config_data.items():
                                        if inherit_key not in config_data[section] and inherit_key != field.alias and inherit_key not in section_names:
                                            config_data[section][inherit_key] = inherit_value
                                match_config_data_to_model(field.type_, config_data[section], root_config_data=root_config_data, parents=parents + [field.alias])
                                section_contents.append(config_data[section])
                            else:
                                raise ConfigError(f"{full_name(parents, field.alias)} refers to section {section} which does not exist.")
                        else:
                            # Not string, assume it's already the section contents
                            section_contents.append(section)
                    config_data[field.alias] = section_contents
                else:
                    match_config_data_to_model(field.type_, config_data[field.alias], root_config_data=root_config_data, parents=parents + [field.alias])
            elif (
                    field.shape in {SHAPE_LIST, SHAPE_TUPLE, SHAPE_TUPLE_ELLIPSIS, SHAPE_ITERABLE, SHAPE_SEQUENCE}
                    or field.type_ in {list, tuple}
            ):
                if isinstance(value, str):
                    config_data[field.alias] = parse_delimited_list(field, value)
            elif (
                    field.shape in {SHAPE_SET}
                    or field.type_ in {set}
            ):
                if isinstance(value, str):
                    config_data[field.alias] = set(parse_delimited_list(field, value))
            elif (
                    field.shape in {SHAPE_FROZENSET}
                    or field.type_ in {frozenset}
            ):
                if isinstance(value, str):
                    config_data[field.alias] = frozenset(parse_delimited_list(field, value))
            elif (
                    field.shape in {SHAPE_DICT, SHAPE_DEFAULTDICT}
                    or field.type_ in {dict}
            ):
                if isinstance(value, str):
                    try:
                        config_data[field.alias] = parse_as_literal_or_json(value)
                    except ValueError as e:
                        raise ConfigError(f"Field {full_name(parents, field.alias)} {e}")
            elif field.shape == SHAPE_SINGLETON:
                if field.type_ == int:
                    config_data[field.alias] = int(config_data[field.alias])
                elif field.type_ == float:
                    config_data[field.alias] = float(config_data[field.alias])
            else:
                pass
                # We'll let pydantic parse it as is







# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['config_wrangler',
 'config_wrangler.config_data_loaders',
 'config_wrangler.config_templates',
 'config_wrangler.types']

package_data = \
{'': ['*']}

install_requires = \
['StrEnum>=0.4.7,<0.5.0', 'pydantic>=1.8.2,<2.0.0']

extras_require = \
{'redshift': ['boto3>=1.4,<2.0'],
 's3': ['boto3>=1.4,<2.0'],
 'sqlalchemy': ['SQLAlchemy>=1.4,<2.0']}

setup_kwargs = {
    'name': 'config-wrangler',
    'version': '0.1.1',
    'description': 'pydantic based configuration wrangler. Handles reading multiple ini or toml files with inheritance rules and variable expansions.',
    'long_description': '# Config Wrangler\n\n[![pypi](https://img.shields.io/pypi/v/config-wrangler.svg)](https://pypi.org/project/config-wrangler/)\n[![license](https://img.shields.io/github/license/arcann/config_wrangler.svg)](https://github.com/arcann/config_wrangler/blob/master/LICENSE)\n\npydantic based configuration wrangler. Handles reading multiple ini or toml files with inheritance rules and variable expansions.\n\n\n\n## Installation\n\nInstall using `pip install -U config-wrangler` or `conda install config-wrangler -c conda-forge`.\n\n## A Simple Example\n\nconfig.ini\n```ini\n[S3_Source]\nbucket_name=my.exmple-bucket\nkey_prefixes=processed/\nuser_id=AK123456789ABC\n# Not a secure way to store the password, but OK for local prototype or examples.\n# See KEYRING or KEEPASS for better options\npassword_source=CONFIG_FILE\nraw_password=My secret password\n\n[target_database]\ndialect=sqlite\ndatabase_name=${test_section:my_environment:source_data_dir}/example_db\n\n[test_section]\nmy_int=123\nmy_float=123.45\nmy_bool=Yes\nmy_str=ABC☕\nmy_bytes=ABCⓁⓄⓋ☕\nmy_list_auto_c=a,b,c\nmy_list_auto_nl=\n    a\n    b\n    c\nmy_list_auto_pipe=a|b|c\nmy_list_c=a,b,c\nmy_list_python=[\'x\',\'y\',\'z\']\nmy_list_json=["J","S","O","N"]\nmy_list_nl=\n    a\n    b\n    c\nmy_list_int_c=1,2,3\nmy_tuple_c=a,b,c\nmy_tuple_nl=\n    a\n    b\n    c\nmy_tuple_int_c=1,2,3\nmy_dict={1: "One", 2: "Two"}\nmy_dict_str_int={"one": 1, "two": 2}\nmy_set={\'A\',\'B\',\'C\'}\nmy_set_int=1,2,3\nmy_frozenset=A,B,C\nmy_date=2021-05-31\nmy_time=11:55:23\nmy_datetime=2021-05-31 11:23:53\nmy_url=https://localhost:6553/\n\n[test_section.my_environment]\nname=dev\n# For example to run we\'ll make both paths relative to current\ntemp_data_dir=.\\temp_data\\${test_section:my_environment:name}\nsource_data_dir=.\n```\n\npython code\n```py\nimport typing\nfrom datetime import date, time, datetime\n\nfrom pydantic import BaseModel, DirectoryPath, Field, AnyHttpUrl\n\nfrom config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader\nfrom config_wrangler.config_from_ini_env import ConfigFromIniEnv, ConfigFromLoaders\nfrom config_wrangler.config_templates.config_hierarchy import ConfigHierarchy\nfrom config_wrangler.config_templates.s3 import S3_Bucket\nfrom config_wrangler.config_templates.sqlalchemy_database import SQLAlchemyDatabase\nfrom config_wrangler.types.paths_types import AutoCreateDirectoryPath\n\n\nclass S3_Bucket_KeyPrefixes(S3_Bucket):\n    key_prefixes: typing.List[str]\n\n\nclass Environment(ConfigHierarchy):\n    name: str = Field(..., env=\'env_name\')\n    temp_data_dir: AutoCreateDirectoryPath\n    source_data_dir: DirectoryPath\n\n\nclass TestSection(BaseModel):\n    my_int: int\n    my_float: float\n    my_bool: bool\n    my_str: str\n    my_bytes: bytes\n    my_list_auto_c: list\n    my_list_auto_nl: list\n    my_list_auto_pipe: list\n    my_list_python: list\n    my_list_json: list\n    my_list_c: list = Field(delimiter=\',\')\n    my_list_nl: list = Field(delimiter=\'\\n\')\n    my_list_int_c: typing.List[int] = Field(delimiter=\',\')\n    my_tuple_c: tuple = Field(delimiter=\',\')\n    my_tuple_nl: tuple = Field(delimiter=\'\\n\')\n    my_tuple_int_c: typing.Tuple[int, int, int] = Field(delimiter=\',\')\n    my_dict: dict\n    my_dict_str_int: typing.Dict[str, int]\n    my_set: set\n    my_set_int: typing.Set[int]\n    my_frozenset: frozenset\n    my_date: date\n    my_time: time\n    my_datetime: datetime\n    my_url: AnyHttpUrl\n    my_environment: Environment\n\n\nclass ETLConfig(ConfigFromIniEnv):\n\n    class Config:\n        validate_all = True\n        validate_assignment = True\n        allow_mutation = True\n\n    target_database: SQLAlchemyDatabase\n\n    s3_source: S3_Bucket_KeyPrefixes\n\n    test_section: TestSection\n\n\nclass ETLConfigAnyLoaders(ETLConfig):\n    def __init__(\n            self,\n            _config_data_loaders: typing.List[BaseConfigDataLoader],\n            **kwargs: typing.Dict[str, typing.Any]\n    ) -> None:\n        # Skip super and call the next higher class\n        ConfigFromLoaders.__init__(\n            self,\n            _config_data_loaders=_config_data_loaders,\n            **kwargs\n        )\n\n\ndef main():\n    config = ETLConfig(file_name=\'simple_example.ini\')\n\n    print(f"Temp data dir = {config.test_section.my_environment.temp_data_dir}")\n    #> Temp data dir = temp_data\\dev\n    \n    print(f"Source data dir = {config.test_section.my_environment.source_data_dir}")\n    #> Source data dir = .\n\n    print(f"my_int = {config.test_section.my_int}")\n    #> my_int = 123\n    \n    print(f"my_float = {config.test_section.my_float}")\n    #> my_float = 123.45\n    \n    print(f"my_str = {config.test_section.my_str}")\n    #> my_str = ABC☕\n\n    print(f"my_list_auto_c = {config.test_section.my_list_auto_c}")\n    #> my_list_auto_c = [\'a\', \'b\', \'c\']\n\n    print(f"my_list_auto_nl = {config.test_section.my_list_auto_nl}")\n    #> my_list_auto_c = [\'a\', \'b\', \'c\']\n\n    print(f"my_dict = {config.test_section.my_dict}")\n    #> my_dict = {1: \'One\', 2: \'Two\'}\n\n    print(f"my_set = {config.test_section.my_set}")\n    #> my_set = {\'C\', \'A\', \'B\'}\n\n    print(f"my_time = {config.test_section.my_time}")\n    #> my_time = 11:55:23\n\n    print(f"my_datetime = {config.test_section.my_datetime}")\n    #> my_datetime = 2021-05-31 11:23:53\n\n    print(f"my_url = {config.test_section.my_url}")\n    #> my_url = https://localhost:6553/\n\n  # Getting DB engine (requires sqlalchemy optional install\n    engine = config.target_database.get_engine()\n    print(f"target_database.engine = {engine}")\n    #> target_database.engine = Engine(sqlite:///.example_db)\n\n    # raise SystemExit("The S3 and DB code below won\'t actually work without an environment specific config")\n\n    print("Getting S3 Data")\n    bucket = config.s3_source.get_bucket()\n    print(f"S3 bucket definition = {bucket}")\n    for prefix in config.s3_source.key_prefixes:\n        print(f"  bucket search prefix = {prefix}")\n    #> Getting S3 Data\n    #> credentials.py:56: UserWarning: Passwords stored directly in config or worse in code are not safe. Please make sure to fix this before deploying.\n    #> S3 bucket definitition = s3.Bucket(name=\'my.exmple-bucket\')\n    #> bucket search prefix = processed/\n\n\nif __name__ == \'__main__\':\n    main()\n\n```\n',
    'author': 'Derek Wood',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/arcann/config_wrangler',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import typing

from pydantic import BaseModel, PrivateAttr

from config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader, EnvConfigDataLoader
from config_wrangler.config_data_loaders.ini_config_data_loader import IniConfigDataLoader
from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_templates.credentials import Credentials, PasswordDefaults
from config_wrangler.utils import merge_configs, interpolate_values


class ConfigFromLoaders(BaseModel):
    """
    Base class for settings, allowing values to be set by files or environment variables.
    """
    passwords: PasswordDefaults = None

    _name_map = PrivateAttr(default={})

    # noinspection PyMethodParameters
    def __init__(
        __pydantic_self__,
        _config_data_loaders: typing.List[BaseConfigDataLoader],
        **kwargs: typing.Dict[str, typing.Any]
    ) -> None:
        log = logging.getLogger(__name__)
        # Uses something other than `self` the first arg to allow "self" as a settable attribute
        config_data = dict(**kwargs)
        for loader in _config_data_loaders:
            loader_config_data = loader.read_config_data(__pydantic_self__)
            merge_configs(config_data, loader_config_data)
        interpolate_values(config_data, config_data)
        config_data = __pydantic_self__.translate_config_data(config_data)
        super().__init__(**config_data)
        errors = set()
        __pydantic_self__.fill_hierarchy(
            model_level=__pydantic_self__,
            parents=[],
            errors=errors,
        )
        if len(errors) > 0:
            indent = ' '*3
            errors_str = f"\n{indent}".join(errors)
            raise ValueError(f"Config Errors (cnt={len(errors)}). Errors=\n{indent}{errors_str}")

    def fill_hierarchy(
            self,
            model_level: BaseModel,
            parents: typing.List[str],
            errors: set,
    ):
        log = logging.getLogger(__name__)
        if len(parents) > 10:
            raise ValueError(f"Possible model self reference {parents}")
        if isinstance(model_level, ConfigHierarchy):
            model_level._root_config = self
            model_level._parents = parents
        for attr_name, attr_value in model_level.__dict__.items():
            if isinstance(attr_value, BaseModel):
                self.fill_hierarchy(
                    model_level=attr_value,
                    parents=parents + [attr_name],
                    errors=errors
                )

            if isinstance(attr_value, Credentials):
                if attr_value.validate_password_on_load:
                    try:
                        _ = attr_value.get_password()
                    except Exception as e:
                        log.error(f"Error: {repr(e)}")
                        errors.add(repr(e))

    # noinspection PyMethodMayBeStatic
    def translate_config_data(self, config_data: typing.MutableMapping):
        return config_data

    def _translate_name(self, old_name):
        if old_name in self._name_map:
            return self._name_map[old_name]
        else:
            return old_name

    def get(self, section, item, fallback=...):
        try:
            section_obj = getattr(self, self._translate_name(section))
            return getattr(section_obj, item)
        except AttributeError:
            if fallback is ...:
                raise
            else:
                return fallback

    def __getitem__(self, section):
        try:
            section_obj = getattr(self, section)
            return section_obj.dict()
        except AttributeError as e:
            raise KeyError(str(e))


class ConfigFromIniEnv(ConfigFromLoaders):
    # noinspection PyMethodParameters
    def __init__(
            __pydantic_self__,
            file_name: str = 'config.ini',
            start_path: typing.Optional[str] = None,
            **kwargs: typing.Dict[str, typing.Any]
    ) -> None:
        env_loader = EnvConfigDataLoader()
        ini_loader = IniConfigDataLoader(start_path=start_path, file_name=file_name)
        super().__init__(
            _config_data_loaders=[env_loader, ini_loader],
            **kwargs
        )



import logging
import os
import typing
from copy import deepcopy

from pydantic import BaseModel

from config_wrangler.config_exception import ConfigError


class BaseConfigDataLoader:
    def __init__(self, config_data_dict: typing.Dict[str, typing.Any] = None, **kwargs):
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        if config_data_dict is None:
            self._init_config_data = dict(**kwargs)
        else:
            self._init_config_data = dict(**kwargs).update(config_data_dict)

    def _apply_types(self, config_data, model: BaseModel) -> typing.MutableMapping:
        for field in model.__fields__.values():
            self.log.debug(field.alias)
            # if field.is_complex():
            #     raise NotImplementedError
            # else:
            #     config_data[field.alias] =
        return config_data

    def read_config_data(self, model: BaseModel) -> typing.MutableMapping:
        return self._apply_types(self._init_config_data)

    def save_config_data(self, config_data: BaseModel):
        raise RuntimeError(f"{self.__class__.__name__} doesn't deal with files")


class EnvConfigDataLoader(BaseConfigDataLoader):
    """
    Copied initial version from Pydantic BaseSettings.
    Made to fit into ConfigLoader approach with nested models.
    """

    def __init__(self, env_prefix: str = None, config_data_dict: typing.Dict[str, typing.Any] = None, **kwargs):
        super().__init__(config_data_dict=config_data_dict, **kwargs)
        self.env_prefix = env_prefix or ''

    def read_config_data(self, model: BaseModel) -> typing.MutableMapping:
        config_data = deepcopy(self._init_config_data)

        env_vars = {k.lower(): v for k, v in os.environ.items()}

        for field in model.__fields__.values():
            env_val: typing.Optional[str] = None
            # Note: field.field_info.extra['env_names'] is set from the 'env' variable on the field or in Config.
            #       See https://pydantic-docs.helpmanual.io/usage/settings/#environment-variable-names
            if 'env_names' in field.field_info.extra:
                for env_name in field.field_info.extra['env_names']:
                    env_name = self.env_prefix + env_name
                    env_val = env_vars.get(env_name.lower())
                    if env_val is not None:
                        break
            if env_val is None:
                for env_name in [field.alias, field.name]:
                    env_name = self.env_prefix + env_name
                    env_val = env_vars.get(env_name.lower())
                    if env_val is not None:
                        break

            if field.is_complex():
                if env_val is not None:
                    try:
                        env_val = config_data.__config__.json_loads(env_val)  # type: ignore
                    except ValueError as e:
                        raise ConfigError(f'error parsing JSON for "{field}"') from e
                else:
                    pass
                    # TODO: Look for nested env var variables by concatenating field names with underscores
                    # DATABASE_HOST
                    # DATABASE_PORT
                    # MY_CONNECTION_HOST_NAME
                    # MY_CONNECTION_PORT

            # Only set the value if it is not already in our config_data
            if env_val is not None and field.alias not in config_data:
                config_data[field.alias] = env_val
        return config_data





import os
from pathlib import Path
from pydantic import DirectoryPath
from pydantic.validators import path_validator

__all__ = ['DirectoryPath', 'AutoCreateDirectoryPath']


class AutoCreateDirectoryPath(DirectoryPath):
    @staticmethod
    def __ensure_exsits(path: Path):
        if not path.exists():
            os.makedirs(path)
        return path

    @classmethod
    def __get_validators__(cls):
        yield path_validator
        yield AutoCreateDirectoryPath.__ensure_exsits
        yield cls.validate
